# -*- coding: utf-8 -*-
# Author: Mikhail Polyanskiy
# Last modified: 2017-04-02
# Original data: Rakić et al. 1998, https://doi.org/10.1364/AO.37.005271

import numpy as np
import matplotlib.pyplot as plt

# Lorentz-Drude (LD) model parameters
# ωp = 9.03  #eV
# f0 = 0.760
# Γ0 = 0.053 #eV
ωp = 9.0175  #eV
f0 = 0.760
Γ0 = 0.02666 #eV

f1 = 0.024
Γ1 = 0.241 #eV
ω1 = 0.415 #eV

f2 = 0.010
Γ2 = 0.345 #eV
ω2 = 0.830 #eV

f3 = 0.071
Γ3 = 0.870 #eV
ω3 = 2.969 #eV

f4 = 0.601
Γ4 = 2.494 #eV
ω4 = 4.304 #eV

f5 = 4.384
Γ5 = 2.214 #eV
ω5 = 13.32 #eV

Ωp = f0**.5 * ωp  #eV

def LD(ω):  #ω: eV
    # ε = 1-Ωp**2/(ω**2)
    ε = 1-Ωp**2/(ω*(ω+1j*Γ0))
    # ε += f1*ωp**2 / ((ω1**2-ω**2)-1j*ω*Γ1)
    # ε += f2*ωp**2 / ((ω2**2-ω**2)-1j*ω*Γ2)
    # ε += f3*ωp**2 / ((ω3**2-ω**2)-1j*ω*Γ3)
    # ε += f4*ωp**2 / ((ω4**2-ω**2)-1j*ω*Γ4)
    # ε += f5*ωp**2 / ((ω5**2-ω**2)-1j*ω*Γ5)
    return ε+0j # forced to be complex-valued

ev_min=0.01
ev_max=15
npoints=200
eV = np.logspace(np.log10(ev_min), np.log10(ev_max), npoints)
μm = 4.13566733e-1*2.99792458/eV*1e-6
ε = LD(eV)
n = (ε**.5).real
k = (ε**.5).imag

#============================   DATA OUTPUT   =================================
file = open('auDrude.ref', 'w')
for i in range(npoints-1, -1, -1):
    file.write('{:.4e} {:.4e} {:.4e} \n'.format(μm[i],n[i],k[i]))
file.close()


#===============================   PLOT   =====================================
plt.rc('font', family='Arial', size='14')

# plt.figure(1)
# plt.plot(eV, -ε.real, label="-ε1")
# plt.plot(eV, ε.imag, label="ε2")
# plt.xlabel('Photon energy (eV)')
# plt.ylabel('ε')
# plt.xscale('log')
# plt.yscale('log')
# plt.legend(bbox_to_anchor=(0,1.02,1,0),loc=3,ncol=2,borderaxespad=0)

# #plot n,k vs eV
#
# plt.figure(2)
# plt.plot(eV, n, label="n")
# plt.plot(eV, k, label="k")
# plt.xlabel('Photon energy (eV)')
# plt.ylabel('n, k')
# plt.yscale('log')
# plt.legend(bbox_to_anchor=(0,1.02,1,0),loc=3,ncol=2,borderaxespad=0)

# plot n,k vs μm
#
plt.figure(3)
plt.plot(μm, n, label="n")
plt.plot(μm, k, label="k")
plt.xlabel('Wavelength (μm)')
plt.ylabel('n, k')
# plt.xscale('log')
# plt.yscale('log')
plt.legend(bbox_to_anchor=(0,1.02,1,0),loc=3,ncol=2,borderaxespad=0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: shelvon
@email: xiaorun.zang@outlook.com

"""

import os
import sys
import time
import resource
import numpy as np
import matplotlib.pyplot as plt

import copy

# For a local debugging, we import or install the local package.
#
# - Method 1, append the folder's path of the local package:

try:
    import ngslab as slab
except:
    srcpath = os.path.dirname(
        os.path.dirname(
            os.path.abspath(os.getcwd())
            )
        )
    sys.path.append(srcpath)
    import ngslab as slab
#
# - Method 2, pip install locally the python package, where "." is intended
# to be run under the root folder of the python package.
# python3 -m pip install --upgrade .
#
# - Method 3, conda install the local package build by conda-build.
# conda install path/to/local/package

# import slab

# print(dir()) # print what have been imported

# sys.exit(0)
#%% 4-layer waveguide:
# a waveguide given in Fig. 2 of Ref. [J. Chilwell and I. Hodgkinson, ‘Thin-films field-transfer matrix theory of planar multilayer waveguides and reflection from prism-loaded waveguides’, J. Opt. Soc. Am. A, vol. 1, no. 7, pp. 742–753, Jul. 1984, doi: 10.1364/JOSAA.1.000742]
def wg_4layer(geom_tSlab):
    nSlab = np.maximum(geom_nPML, round(np.ceil(geom_tSlab/geom_hMax)))

    wg = dict(
        # geometry
        name = "demo_4layer",
        intervals = [-4*geom_tSlab-geom_tSub-geom_tPML, -4*geom_tSlab-geom_tSub, -4*geom_tSlab, -3*geom_tSlab, -2*geom_tSlab, -geom_tSlab, 0, geom_tFree, geom_tFree+geom_tPML],
        labels = ["pml_left", "sub", "slab4","slab3","slab2","slab1", "air", "pml_right"],
        nnodes = [geom_nPML, round(np.ceil(geom_tSub/geom_hMax)), nSlab, nSlab, nSlab, nSlab, round(np.ceil(geom_tFree/geom_hMax)), geom_nPML],
        # material
        nk = {"default":1.0, "nc": 1.0, "ns": 1.5, "n1": 1.66, "n2": 1.53, "n3": 1.60, "n4": 1.66},
        map = {"pml_left": "ns", "sub":"ns", "slab4":"n4", "slab3":"n3", "slab2":"n2", "slab1":"n1", "air":"nc", "pml_right":"nc"},
        )

    return wg

#### 4-layer lossy waveguide:
def wg_4layer_lossy(geom_tSlab):
    wg = wg_4layer(geom_tSlab)

    wg.update(
        name="demo_4layer_lossy",
        nk={"default":1.0, "nc": 1.0, "ns": 1.5, "n1": 1.66*(1+.0001j), "n2": 1.53*(1+.0001j), "n3": 1.60, "n4": 1.66},
        map = {"pml_left": "ns", "sub":"ns", "slab4":"n4", "slab3":"n3", "slab2":"n2", "slab1":"n1", "air":"nc", "pml_right":"nc"},
        )

    return wg

#### ARROW waveguide:
# in Fig. 2 of Ref. [H. P. Uranus, H. J. W. M. Hoekstra, and E. Van Groesen, ‘Simple high-order Galerkin finite element scheme for the investigation of both guided and leaky modes in anisotropic planar waveguides’, Optical and Quantum Electronics, vol. 36, no. 1, pp. 239–257, Jan. 2004, doi: 10.1023/B:OQEL.0000015643.52433.f0.]
def wg_ARROW():
    d1 = 2.0985e-6*1e6;
    d2 = 0.1019e-6*1e6;
    d3 = 4e-6*1e6;
    nSlab = np.maximum(geom_nPML, round(np.ceil(geom_tSlab/geom_hMax)));
    wg = dict(
        name = "demo_ARROW",
        intervals = [-d1-d2-d3-geom_tSub-geom_tPML, -d1-d2-d3-geom_tSub, -d1-d2-d3, -d2-d3, -d3, 0, geom_tFree, geom_tFree+geom_tPML],
        labels = ["pml_left", "sub", "slab1","slab2","slab3", "air", "pml_right"],
        nnodes = [geom_nPML, round(np.ceil(geom_tSub/geom_hMax)), nSlab, nSlab, nSlab, round(np.ceil(geom_tFree/geom_hMax)), geom_nPML],
        nk = {"default":1.0, "nc": 1.0, "ns": 3.5, "n1": 1.45, "n2": 3.5, "n3": 1.45},
        map = {"pml_left": "ns", "sub":"ns", "slab1":"n1", "slab2":"n2", "slab3":"n3", "air":"nc", "pml_right":"nc"},
        )

    return wg

#%% model creation
#---- Geometric Parameters
model = slab.SlabWaveGuide();
model.ld0_target = 6328e-10

# use micrometer may yield lower condition number
# model.ld0_target *= 1e6

geom_tFree = model.ld0_target*0.2
geom_tSub = model.ld0_target*0.2
geom_tPML = model.ld0_target*0.2

# geom_hMax = model.ld0_target/31; geom_nPML = 13; # extra fine mesh
# geom_hMax = model.ld0_target/31; geom_nPML = 27; # extra fine PML mesh
# geom_hMax = model.ld0_target/17; geom_nPML = 13; # moderate resolution
# geom_hMax = model.ld0_target/31; geom_nPML = 1; # TBC
geom_hMax = model.ld0_target/17; geom_nPML = 1; # TBC debugging

geom_delPML = geom_tPML/geom_nPML
print("hMax="+f"{geom_hMax/model.ld0_target:.4f}*ld0")
print("hPML="+f"{geom_delPML/model.ld0_target:.4f}*ld0")

#%% loop in the slab thickness
tSlabArray = np.arange(500e-9, 501e-9, 20e-9)

# Here, a change of unit to micrometer is needed as well.
# tSlabArray *= 1e6
for it, geom_tSlab in zip(range(tSlabArray.size), tSlabArray):
    print("tSlab["+str(it)+"] = "+f"{geom_tSlab*1e3:.0f} nm")

    # choose a waveguide
    # wg = wg_4layer(geom_tSlab)
    wg = wg_4layer_lossy(geom_tSlab)
    # wg = wg_ARROW()

    # update the waveguide geometry
    model.geom.Update(wg["intervals"], wg["nnodes"], wg["labels"])

    #-- create the mesh
    model.mesh.Create()
    # model.mesh.Plot(); #sys.exit(0);

    model.material.nk = wg["nk"]
    model.material.map = wg["map"]

    # enable PML domains
    # model.material.pml = {"pml_left":True, "pml_right":True}

    # disable PML
    # model.material.pml = {"pml_left":False, "pml_right":False} # False by default

    # and then set the transparent boundary condition
    # setTBC() accepts two numbers defining the indices of the leftmost to the rightmost semi-infinite regions that bounding the core waveguide region
    model.setTBC(tbc_doms=[1, -2]) # by default: tbc_doms=[1, -2]
    # model.setTBC(tbc_doms=[0, -1]) # test

    # model.mesh.Plot(); # sys.exit(0);
    # model.TBC_PEP = False

    # sys.exit(0)
    #%% model build
    #---- build the model (create function space, set up materials, enable pml materials)
    # bspl_order: the order of the ngsolve.BSpline representations for dispersive and tabulated materials
    model.Build(fes_order=2, bspl_order=2, ld_scale=1) # bspl_order = 1 (linear), 2 (quadratic) curve

    # debug PML
    # model.SetPML(model.material.pml, pml_plot=True);

    # sys.exit(0)
    #%% model simulation

    #---- Sweep normalized frequency and solve the problem
    model.wArray = np.array([1.0])

    # model.wArray = np.concatenate( (model.wArray, np.arange(0.6, 2.8, 0.1)) )

    model.wArray = np.unique(np.round(model.wArray*1000))/1000 # keep unique floating numbers of 3 decimal precision.

    # figure properties
    plt.close("all")

    inch2cm = 1.0/2.54  # inches-->centimeters
    figsize = 1.0*np.array([8.6, 6.4])*inch2cm
    ms = 4
    fontsize = 8
    plt.rc('font', size=fontsize)

    fig, axsSpectra = plt.subplots(figsize=figsize, ncols=2, nrows=1, constrained_layout=True, sharex=True, sharey=True);
    # fig, axsSpectra = plt.subplots(figsize=figsize, ncols=1, nrows=1, constrained_layout=True, sharex=True, sharey=True);
    ll_plotted = False

    # print("ldArray="+str(model.ld0_target/model.wArray)+" um")
    for iw in range(model.wArray.size):
    # for iw in range(model.wArray.size)[::-1]: # in reverse order
        model.w = model.wArray[iw]

        t0 = time.process_time()
        sol = model.Solve(show_pattern=False)
        t1 = time.process_time()
        t_CPU = t1 - t0

        print('CPU Execution time:', t_CPU, 'seconds')
        # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        # print('Memory usage:', mem/1024/8, 'MB') # not clear the meaning of this parameter.


        #%% investigate the eigenvalues
        cmaps_r = ["C7", "C3", "C2", "C5"] # complementary color of each
        c_fem = ["g", "C7"]
        alpha_fem = [1.0, 0.6]
        c_tmatrix = ["C3"]

        markerstyle_tmatrix = dict(marker="x", s=ms*3, linewidths=0.5, alpha=1.0, clip_on=True)#, zorder=1e3
        markerstyle_fem = dict(marker="o", s=round(ms*2), edgecolor="none", clip_on=True)#, zorder=1e6

        n2list = np.asarray(model.sol.n2list)
        nlist = np.asarray(n2list)**0.5
        n2delta = n2list[-1] - n2list[0]
        n2Sigma = n2list[-1] + n2list[0]
        ns, nc = nlist[[0, -1]]
        n2s, n2c = nlist[[0, -1]]**2

        #### filtering modes
        mode_markers = ["o", "^"]
        Q_threshold = 1
        # Q_threshold = (1/np.e)
        # Q_threshold = 1e-1
        kappa_tol = 1e-12
        kappa2_limit = np.max(np.abs(n2list))*100.1

        # the last one will be plotted in the complex planes
        # mode_type == 0: TM mode; mode_type == 1: TE mode
        for mode_type in [0]:
        # for mode_type in [1, 0]:
        # for mode_type in [0, 1]:

            ## results from FEM simulations
            nvalid = model.sol._nvalid[mode_type]
            kappa_fem = model.sol.kappa[mode_type, :nvalid]
            kappa2_fem = model.sol.kappa2[mode_type, :nvalid]
            taus_fem = model.sol.taus[mode_type, :nvalid]
            tauc_fem = model.sol.tauc[mode_type, :nvalid]

            idx_kappa_fem = np.where(
                ~np.isnan(kappa_fem)
                & (np.abs(kappa_fem.real/kappa_fem.imag)>Q_threshold)
                & (np.abs(kappa_fem)**2<kappa2_limit)
                )[0]

            # divided tau plane into regions by the sign of Sx
            if mode_type==0: # TM mode
                Sx_s = (taus_fem.real*n2s.real+taus_fem.imag*n2s.imag)
                Sx_c = (tauc_fem.real*n2c.real+tauc_fem.imag*n2c.imag)
            elif mode_type==1: # TE mode for nonmagnetic material with (mu.real = 1, mu.imag=0)
                Sx_s = -(taus_fem.real)
                Sx_c = -(tauc_fem.real)

            # Keep modes that do not have energy feeding into the waveguide region from the infinity
            # That means, the situations of both the power flow direction points into the waveguide
            # and density increases away from the waveguide cannot be met simultaneously.
            #
            #
            #FIXME: According to the following paper
                # [1] T. Tamir and F. Kou, ‘Varieties of leaky waves and their excitation along multilayered structures’, IEEE Journal of Quantum Electronics, vol. 22, no. 4, pp. 544–551, Apr. 1986, doi: 10.1109/JQE.1986.1072991.
                # The contra-leaky mode V cannot be excited by any realistic sources.
                # However, this argument is not so obvious without giving any further explanation in that paper.
            idx_s = np.where(
                ( ~((Sx_s>-kappa_tol) & (taus_fem.imag>-kappa_tol)) )
                )[0]
            idx_c = np.where(
                ( ~((Sx_c<kappa_tol) & (tauc_fem.imag<kappa_tol)) )
                )[0]

            # mode indices of power flux from the denser cladding layer to the rarer cladding layer
            if ns.real>=nc.real:
                idx_denser2rarer = np.where(
                    (~(
                        ((Sx_s>-kappa_tol) & (taus_fem.imag<kappa_tol)) # power flow type As
                        &((Sx_c>-kappa_tol) & (tauc_fem.imag<kappa_tol)) # power flow type Cc
                       ))
                    )[0]

            idx = np.intersect1d(idx_s, idx_c)
            # idx = np.intersect1d(idx, idx_denser2rarer)
            idx = np.intersect1d(idx_kappa_fem, idx)
            # idx = idx_denser2rarer

            # idx = idx_kappa_fem
            idx_del = np.setdiff1d(np.arange(nvalid), idx)

            if idx.size>0:

                # modes selected by all three conditions
                c = range(idx.size)
                # c = "C0"
                axsSpectra[mode_type].scatter(np.abs(kappa_fem[idx].real)*model.k0, np.repeat(model.w, idx.size), c=c, cmap="seismic", s=8, alpha=np.exp(-np.abs(kappa2_fem[idx].imag)-kappa_tol)**0)

                # modes selected by only the condition on kappa
                # axsSpectra[mode_type].scatter(np.abs(kappa_fem[mode_type, idx_kappa_fem].real)*model.k0, np.repeat(model.w, idx_kappa_fem.size), c="C7", s=4, alpha=np.exp(-np.abs(kappa2_fem[mode_type,idx].imag)-kappa_tol)**0)
                # mode w/o any selection
                # axsSpectra[mode_type].scatter(kappa_fem[mode_type, idx_del].real*model.k0, np.repeat(model.w, idx_del.size), c="C7", s=4)

                ## complex refractive index of a dispersive medium
                # axsSpectra[0].scatter(nlist[1].real*2*np.pi/model.ld0, model.w, c="b")
                # axsSpectra[0].scatter(nlist[1].imag*2*np.pi/model.ld0, model.w, c="g")

            # axsSpectra.set_xlim([0, 250])
            axsSpectra[mode_type].set_xlim([0, 250])

        # plot two light lines once
        if not ll_plotted:

            axsSpectra[0].axline((0,0), (nc*2*np.pi/model.ld0Array[0], model.wArray[0]), c="k")
            axsSpectra[1].axline((0,0), (nc*2*np.pi/model.ld0Array[0], model.wArray[0]), c="k")

            axsSpectra[0].axline((0,0), (ns*2*np.pi/model.ld0Array[0], model.wArray[0]), c="gray")
            axsSpectra[1].axline((0,0), (ns*2*np.pi/model.ld0Array[0], model.wArray[0]), c="gray")

            ll_plotted = True

        #### at a single frequency
        # investigate eigenvalues in the complex planes
        if (model.w==1.0):

            sol_probe = copy.deepcopy(sol)

            #### results from the t-matrix calculations
            if wg["name"]=="demo_4layer":
                neff_tmatrix = np.genfromtxt("4layer_tmatrix.txt", dtype=np.complex128)
                neff_tmatrix = neff_tmatrix.reshape((2, int(neff_tmatrix.size/2)))
            elif wg["name"]=="demo_4layer_lossy":
                neff_tmatrix = np.genfromtxt("4layer_lossy_tmatrix.txt", dtype=np.complex128)
                neff_tmatrix = neff_tmatrix.reshape((2, int(neff_tmatrix.size/2)))
            else:
                neff_tmatrix = np.full((2,1), np.nan)

            nmodes_plot = np.min([12, np.size(neff_tmatrix, 1)])
            kappa_tmatrix = neff_tmatrix[:,:nmodes_plot]
            kappa2_tmatrix = kappa_tmatrix**2

            # plus a negligible imag to get the correct sign when it is pure real
            tau2_s_tmatrix = np.array(n2list[0]-kappa2_tmatrix) - 1j*1e-32
            tau2_c_tmatrix = np.array(n2list[-1]-kappa2_tmatrix) - 1j*1e-32

            taus_tmatrix = -np.sqrt(tau2_s_tmatrix)*np.sign(np.real(tau2_s_tmatrix))
            tauc_tmatrix = np.sqrt(tau2_c_tmatrix)*np.sign(np.real(tau2_c_tmatrix))

            eig_tmatrix = 0.5*(tauc_tmatrix - taus_tmatrix)

            plt.close("all")

            #### plot three regions in the complex planes of kappa, lambda, taus, and tauc
            fig, _axs = plt.subplots(figsize = figsize, ncols = 2, nrows = 2, constrained_layout = False);
            axs = _axs.flatten()

            slab.plotRegions(axs, nklist=nlist)
            # slab.plotRegions(axs, nklist=nlist, kappa_modes=kappa_tmatrix[mode_type,:])

            # cmap_fem = "viridis_r"
            cmap_fem = "turbo"
            cmap_fem_del = "Grays"

            axsScatter = []
            for ax, zz, zz_tmatrix in zip([axs[0], axs[1], axs[2], axs[3]],
                              [sol_probe.kappa[mode_type, :sol_probe._nvalid[mode_type]],
                               sol_probe._eigval[mode_type, :sol_probe._nvalid[mode_type]],
                               sol_probe.taus[mode_type, :sol_probe._nvalid[mode_type]],
                               sol_probe.tauc[mode_type, :sol_probe._nvalid[mode_type]]],
                              [kappa_tmatrix[mode_type,:],
                               eig_tmatrix[mode_type,:],
                               taus_tmatrix[mode_type,:],
                               tauc_tmatrix[mode_type,:]]
                              ):

                ax.scatter(zz[idx_del].real, zz[idx_del].imag, c=c_fem[1], alpha=alpha_fem[0], **markerstyle_fem)
                # markerstyle_fem["zorder"] += 1
                axsScatter.append(ax.scatter(
                    zz[idx].real, zz[idx].imag,
                    c=range(idx.size)[::-1],
                    cmap=cmap_fem,
                    alpha=alpha_fem[0],
                    **markerstyle_fem))

                ax.scatter(zz_tmatrix.real, zz_tmatrix.imag, c=c_tmatrix[0], **markerstyle_tmatrix)

            # extract the field profile of selected modes
            axs_names = ["kappa", "eigval", "taus", "tauc"]
            axs_select = np.asarray(axsScatter)

            # plt.close(1) # close the figure for plotting the dispersion curve
            ####  interactively select a mode by clicking at the nearest point in the complex plane
            mode_selector = slab.SlabWaveGuide.modeSelector(
                axs_select, sol_probe, axs_names, mode_type=mode_type, nmodes = 2,
                )

            kappa_max = np.max(np.abs(np.asarray(sol_probe.nlist)))
            #### plot limits for the 4layer, lossless waveguide case
            # continue
            if wg["name"]=="demo_4layer" or wg["name"]=="demo_4layer_lossy":
                axs[0].set_xlim([0.21, kappa_max*1.02])
                axs[0].set_ylim([-0.02, 0.67])

                axs[0].text(sol_probe.nlist[0].real+0.03, 0.32, '$n_{\mathrm{s}}$', usetex=True, fontsize=fontsize*1.25, ha="left", va="center", fontweight="book")
                axs[0].text(sol_probe.nlist[-1].real+0.03, 0.32, '$n_{\mathrm{c}}$', usetex=True, fontsize=fontsize*1.25, ha="left", va="center", fontweight="book")
                # inset showing a zoom in area
                # if wg["name"]=="demo_4layer_lossy":
                # axs[0].set_ylim([-0.02, 0.67])
                x1, x2, y1, y2 = sol_probe.nlist[0].real*0.99, kappa_max*0.99, -0.0001, 0.0003  # subregion of the original image
                axins = axs[0].inset_axes(
                    [0.3, 0.75, 0.8, 0.35],
                    xlim=(x1, x2), ylim=(y1, y2), xticks=[], yticks=[])
                zz = sol_probe.kappa
                zz_tmatrix = kappa_tmatrix
                axins.scatter(zz[mode_type,idx_del].real, zz[mode_type,idx_del].imag, c=c_fem[1], alpha=alpha_fem[1], **markerstyle_fem)
                # markerstyle_fem["zorder"] += 1
                axins.scatter(zz[mode_type,idx].real, zz[mode_type,idx].imag, c=range(idx.size)[::-1], cmap=cmap_fem, alpha=alpha_fem[0], **markerstyle_fem)
                axins.scatter(zz_tmatrix[mode_type,:].real, zz_tmatrix[mode_type,:].imag, c=c_tmatrix[0], **markerstyle_tmatrix)
                axins.axline((0, 0), (1,0), linestyle="--", linewidth=1, c="k")

                axs[0].indicate_inset_zoom(axins, edgecolor="k")

                axs[1].set_xlim([-1.35, 1.35])
                axs[1].set_ylim([-1.05, 1.05])
                axs[2].set_xlim([-1.75, 1.75])
                axs[2].set_ylim([-0.75, 0.75])
                axs[3].set_xlim([-1.45, 1.45])
                axs[3].set_ylim([-1.75, 1.75])


                # if wg["name"]=="demo_4layer_lossy":

                #     axs[2].set_xlim([-0.00110, 0.00110])
                #     axs[3].set_xlim([-0.00030, 0.00030])
                #     axs[2].set_ylim([-0.75, 0.75])
                #     axs[3].set_ylim([-1.45, 1.45])

                #     axs[2].set_xscale('symlog')
                #     axs[3].set_xscale('symlog')
                #     axs[2].set_xticks([-0.001, 0, 0.001])
                #     axs[3].set_xticks([-0.0002, 0, 0.0002])

            plt.subplots_adjust(hspace=0.5)
            plt.subplots_adjust(wspace=0.4)
    sys.exit(0)
    # continue
    #%% save the figure
    figname = wg["name"]
    # plt.savefig(figname + '.png',format='png', dpi=300)
    plt.savefig(figname + '.pdf',format='pdf')
    # from PIL import Image
    # Image.open(figname+'.png').convert('L').save(figname+'-bw.png')



# Plot and write dispersive refractive indices of gold,
# using Eq. (1) in Ref. [1].
# [1] Rakić et al. 1998, https://doi.org/10.1364/AO.37.005271

import numpy as np
import matplotlib.pyplot as plt
import scipy.constants as const

# Drude model parameters for silver

omega_p_eV = 9.01  #eV
gamma_eV = 0.048 #eV

eV2omega = 2*np.pi*const.electron_volt/const.Planck
omega_p = eV2omega*omega_p_eV
gamma = eV2omega*gamma_eV
# omega_p = 1.3688599705386422e+16
# gamma = 72924837498174.06

def DrudeAg(omega):
    epsilon = 1-omega_p**2/(omega*(omega+1j*gamma))
    return epsilon+0j # forced to be complex-valued

omega_min = 0.01*omega_p
omega_max = 2.4*omega_p
npoints=200
omega = np.linspace(omega_min, omega_max, npoints)
ld = 2*np.pi*299792458/omega
epsilon2 = 1
omega_spp = omega_p/np.sqrt(1+epsilon2)
ld_spp = 2*np.pi*299792458/omega_spp
ld_p = 2*np.pi*299792458/omega_p
epsilon = DrudeAg(omega)
n = (epsilon**.5).real
k = (epsilon**.5).imag

# save to a file
file = open('agDrude.ref', 'w')
for i in range(npoints-1, -1, -1):
    file.write('{:.4e} {:.4e} {:.4e} \n'.format(ld[i], n[i], k[i]))
file.close()

# plot n,k vs μm
#
plt.rc('font', family='Arial', size='14')
#
plt.figure(3)
plt.plot(ld, n, label="n")
plt.plot(ld, k, label="k")
plt.xlabel('Wavelength (μm)')
plt.ylabel('n, k')
# plt.xscale('log')
# plt.yscale('log')
plt.legend(bbox_to_anchor=(0,1.02,1,0),loc=3,ncol=2,borderaxespad=0)



#!/usr/bin/env python3

"""

@author: shelvon
@email: xiaorun.zang@outlook.com

"""

import numpy as np
import scipy.sparse as scipy_sparse

from petsc4py import PETSc

# convert ngsolve matrix to PETSc matrices
def ngs2petscMatAIJ(ngs_mat):
    locmat = ngs_mat.local_mat
    eh, ew = locmat.entrysizes
    val,col,ind = locmat.CSR()
    ind = np.array(ind).astype(PETSc.IntType)
    col = np.array(col).astype(PETSc.IntType)
    apsc_loc = PETSc.Mat().createBAIJ(size=(eh*locmat.height, eh*locmat.width), bsize=eh, csr=(ind,col,val))

    return apsc_loc

# directly convert ngsolve matrix to numpy matrix
def ngs2numpyMat(ngs_mat):
    rows,cols,vals = ngs_mat.mat.COO()
    np_mat = scipy_sparse.csr_matrix((vals,(rows,cols))).todense()

    return np_mat

# investigate elements of the assembled matrices
def petscMat2numpyMat(v):
    s=v.getValues(range(0, v.getSize()[0]), range(0,  v.getSize()[1]))
    return s



#!/usr/bin/env python3
"""
@author: shelvon
@email: xiaorun.zang@outlook.com

"""

import os
import sys
import time
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.gridspec import GridSpec

import copy

# For a local debugging, we import or install the local package.
#
# - Method 1, append the folder's path of the local package:

try:
    import ngslab as slab
except:
    srcpath = os.path.dirname(
        os.path.dirname(
            os.path.abspath(os.getcwd())
            )
        )
    sys.path.append(srcpath)
    import ngslab as slab
#
# - Method 2, pip install locally the python package, where "." is intended
# to be run under the root folder of the python package.
# python3 -m pip install --upgrade.
#
# - Method 3, conda install the local package build by conda-build.
# conda install path/to/local/package

# import slab

# print(dir()) # print what have been imported

# sys.exit(0)
#%% silver-silica slab waveguide:
def wg_AgSiO2(geom_tSlab):
    nSlab = np.maximum(geom_nPML, round(np.ceil(geom_tSlab/geom_hMax)))
    wg = dict(
        # geometry
        name = "demo_AgSiO2",
        intervals = [-(geom_tPML + geom_tSub + geom_tSlab), -(geom_tSub + geom_tSlab), -geom_tSlab, 0.0, geom_tFree, geom_tFree + geom_tPML],
        labels = ["pml_left", "sub", "slab", "cover", "pml_right"],
        nnodes = [geom_nPML, round(np.ceil(geom_tSub/geom_hMax)), round(np.ceil(geom_tSlab/geom_hMax)), round(np.ceil(geom_tFree/geom_hMax)), geom_nPML],
        # material
        nk = {"default":1.0, "air": 1.0, "SiO2": 1.4537, "Si3N4": 2.0255, "silver": "agjc", "silverDrude":"agDrude"},
        map = {"pml_left":"silver", "sub":"silver", "slab":"SiO2"},
        # map = {"pml_left":"silverDrude", "sub":"silverDrude", "slab":"SiO2"},
        )

    return wg

#%% model creation
#---- Geometric Parameters
model = slab.SlabWaveGuide();
# the wavelength of the angular plasma frequency
# for silver is
# 2*np.pi*299792458/omega_p, where omega_p = 1.3688599705386422e+16 (rad/s)
model.ld0_target = 1.376073234552722e-07

# use micrometer may yield lower condition number and better accuracy
# However, be careful in obtaining wavenumber in the post-processing.
# model.ld0_target *= 1e6

# The thicknesses geom_tFree and geom_tSub become irrelevant
# to numerical accuracy, if TBCs are used.
geom_tFree = model.ld0_target*0.3
geom_tSub = model.ld0_target*0.3
geom_tPML = model.ld0_target*0.2

# geom_hMax = model.ld0_target/31; geom_nPML = 13; # extra fine mesh
# geom_hMax = model.ld0_target/31; geom_nPML = 27; # extra fine PML mesh
geom_hMax = model.ld0_target/17; geom_nPML = 1; # TBC debugging

geom_delPML = geom_tPML/geom_nPML
print("hMax="+f"{geom_hMax/model.ld0_target:.4f}*ld0")
print("hPML="+f"{geom_delPML/model.ld0_target:.4f}*ld0")

#%% loop in the slab thickness
tSlabArray = np.arange(200e-9, 201e-9, 20e-9)

# Here, a change of unit to micrometer is needed as well.
# tSlabArray *= 1e6
for it, geom_tSlab in zip(range(tSlabArray.size), tSlabArray):
    print("tSlab["+str(it)+"] = "+f"{geom_tSlab*1e3:.0f} nm")

    # choose a waveguide
    wg = wg_AgSiO2(geom_tSlab)

    # update the waveguide geometry
    model.geom.Update(wg["intervals"], wg["nnodes"], wg["labels"])

    #-- create the mesh
    model.mesh.Create()
    # model.mesh.Plot();

    model.material.nk = wg["nk"]
    model.material.map = wg["map"]

    # enable PML domains
    # model.material.pml = {"pml_left":True, "pml_right":True}
    # disable PML
    # model.material.pml = {"pml_left":False, "pml_right":False} # False by default

    # and then set the transparent boundary condition
    # setTBC() accepts two numbers defining the indices of the leftmost to the rightmost semi-infinite regions that bounding the core waveguide region
    model.setTBC(tbc_doms=[1, -2]) # by default: tbc_doms=[1, -2]
    # model.setTBC(tbc_doms=[0, -1]) # test

    # model.mesh.Plot()
    # sys.exit(0)
    #%% model build
    #---- build the model (create function space, set up materials, enable pml materials)
    # bspl_order: the order of the ngsolve.BSpline representations for dispersive and tabulated materials
    # model.Build(fes_order=2, bspl_order=2, ld_scale=1e6) # bspl_order = 1 (linear), 2 (quadratic) curve
    model.Build(fes_order=2, bspl_order=2, ld_scale=1) # ld_scale = 1 is no scaling in the dimensions and wavelength

    # debug PML
    # model.SetPML(model.material.pml, pml_plot=True);

    # sys.exit(0)

    #%% figure properties
    plt.close("all")
    inch2cm = 1.0/2.54  # inches-->centimeters
    figsize = 2.0*np.array([8.6, 5.4])*inch2cm
    ms = 4
    fontsize = 8
    plt.rc('font', size=fontsize)

    ll_plotted = False # plot light lines
    c_fem = ["g", "C7"]
    alpha_fem = [1.0, 0.6]
    markerstyle_fem = dict(marker="o", s=round(ms*2), edgecolor="none", clip_on=True)# ,zorder=1)

    fig = plt.figure(figsize = figsize, layout="constrained")
    gs = GridSpec(4, 6, figure=fig)

    axSpectra = fig.add_subplot(gs[0:2, 0:2])
    ax_hz1 = fig.add_subplot(gs[0, 2:4])
    ax_hz2 = fig.add_subplot(gs[0, 4:6])
    ax_hz3 = fig.add_subplot(gs[1, 2:4])
    ax_hz4 = fig.add_subplot(gs[1, 4:6])
    ax_kappa = fig.add_subplot(gs[2:4, 0:2])
    ax_taus = fig.add_subplot(gs[2:4, 2:4])
    ax_tauc = fig.add_subplot(gs[2:4, 4:6])

    axs = [axSpectra, ax_hz1, ax_hz2, ax_hz3, ax_hz4, ax_kappa, ax_taus, ax_tauc]
    for i in range(len(axs)): # create (a), (b), ... labels for each subplot
        x = -0.03
        y = 1.08
        if i==0:
            y = 1.03
        axs[i].text(x, y, '('+chr(i+97)+')', transform=axs[i].transAxes,
                    fontsize=fontsize*1.25, ha="right", va="bottom", fontweight="book", fontname="Sans")

    # sys.exit(0)
    #%% model simulation

    #-- normalized frequencies
    model.wArray = np.array([0.37])
    model.wArray = np.concatenate( (model.wArray, np.arange(0.1, 0.5, 0.01)) )
    model.wArray = np.concatenate( (model.wArray, np.arange(0.3, 0.5, 0.005)) )
    model.wArray = np.unique(np.round(model.wArray*1000))/1000 # keep unique floating numbers of 3 decimal precision.

    t0 = time.process_time()
    # print("ldArray="+str(model.ld0_target/model.wArray)+" um")
    for iw in range(model.wArray.size):
    # for iw in range(model.wArray.size)[::-1]: # in reverse order
        model.w = model.wArray[iw]

        sol = model.Solve(show_pattern=False)

        # continue
        # sys.exit(0)
        #%% investigate the eigenvalues

        n2list = np.asarray(model.sol.n2list)
        nlist = np.asarray(n2list)**0.5
        ns, nc = nlist[[0, -1]]
        n2s, n2c = n2list[[0, -1]]

        #### filtering modes
        mode_markers = ["o", "^"]
        Q_threshold = 1
        # Q_threshold = (1/np.e)
        # Q_threshold = 1e-1
        kappa_tol = 1e-12
        kappa2_limit = np.max(np.abs(n2list))*100.1

        # the last one will be plotted in the complex planes
        # mode_type == 0: TM mode; mode_type == 1: TE mode
        for mode_type in [0]:
        # for mode_type in [1, 0]:
        # for mode_type in [0, 1]:

            ## results from FEM simulations
            nvalid = model.sol._nvalid[mode_type]
            kappa_fem = model.sol.kappa[mode_type, :nvalid]
            kappa2_fem = model.sol.kappa2[mode_type, :nvalid]
            taus_fem = model.sol.taus[mode_type, :nvalid]
            tauc_fem = model.sol.tauc[mode_type, :nvalid]

            # filter on kappa
            idx_kappa_fem = np.where(
                ~np.isnan(kappa_fem)
                & (np.abs(kappa_fem.real/kappa_fem.imag)>Q_threshold)
                & (np.abs(kappa_fem)**2<kappa2_limit)
                )[0]

            # divided tau plane into regions according to the sign of Sx
            if mode_type==0: # TM mode
                Sx_s = (taus_fem.real*n2s.real+taus_fem.imag*n2s.imag)
                Sx_c = (tauc_fem.real*n2c.real+tauc_fem.imag*n2c.imag)
            elif mode_type==1: # TE mode for nonmagnetic material with (mu.real = 1, mu.imag=0)
                Sx_s = -(taus_fem.real)
                Sx_c = -(tauc_fem.real)

            # Keep modes that do not have energy feeding into the waveguide region from the infinity
            # That means, the situations of both the power flow direction points into the waveguide
            # and density increases away from the waveguide cannot be met simultaneously.
            #FIXME: According to the following paper
                # [1] T. Tamir and F. Kou, ‘Varieties of leaky waves and their excitation along multilayered structures’, IEEE Journal of Quantum Electronics, vol. 22, no. 4, pp. 544–551, Apr. 1986, doi: 10.1109/JQE.1986.1072991.
                # The contra-leaky mode V cannot be excited by any realistic sources.
                # However, this argument is not so obvious without giving any further explanation in that paper.
            idx_s = np.where(
                ( ~((Sx_s>-kappa_tol) & (taus_fem.imag>-kappa_tol)) )
                )[0]
            idx_c = np.where(
                ( ~((Sx_c<kappa_tol) & (tauc_fem.imag<kappa_tol)) )
                )[0]

            # mode indices of power flux from the denser cladding layer to the rarer cladding layer
            if ns.real>=nc.real:
                idx_denser2rarer = np.where(
                    (~(
                        ((Sx_s>-kappa_tol) & (taus_fem.imag<kappa_tol)) # power flow type As
                        &((Sx_c>-kappa_tol) & (tauc_fem.imag<kappa_tol)) # power flow type Cc
                       ))
                    )[0]

            idx = np.intersect1d(idx_s, idx_c)
            # idx = np.intersect1d(idx, idx_denser2rarer)
            idx = np.intersect1d(idx_kappa_fem, idx)
            # idx = idx_denser2rarer

            # idx = idx_kappa_fem
            idx_del = np.setdiff1d(np.arange(nvalid), idx)

            if idx.size>0:

                # modes selected by all three conditions
                axSpectra.scatter(
                    np.abs(kappa_fem[idx].real)*model.k0, np.repeat(model.w, idx.size),
                    # c = "C0",
                    c=np.full(idx.size, 1),
                    # c=range(idx.size),
                    # cmap="viridis_r",
                    cmap="seismic",
                    s=ms,
                    alpha=np.exp(-np.abs(kappa2_fem[idx].imag)-kappa_tol)**0
                    )

                # dispersive light line(s) is plotted wavelength by wavelength
                # axSpectra.scatter(np.abs(nlist[0].real)*model.k0, model.w, s=1, c = "gray")

        # plot two light lines once
        if not ll_plotted:

            axSpectra.axline((0,0), (nlist[1]*2*np.pi/model.ld0Array[0], model.wArray[0]), c="gray")
            axSpectra.axline((0,0), (nlist[2]*2*np.pi/model.ld0Array[0], model.wArray[0]), c="gray")

            ll_plotted = True

        # continue
        #### at a single frequency
        # investigate eigenvalues in the complex planes
        if (
            model.w==0.37
            # model.w==0.25
            ):
            sol_probe = copy.deepcopy(sol)

            cmap_fem = "viridis_r"
            cmap_fem_del = "Grays"
            c_hz = ["C0", "C1", "C3", "C9"]

            axsScatter = []
            for ax, zz in zip([ax_kappa, ax_taus, ax_tauc],
                              [sol_probe.kappa[mode_type, :sol_probe._nvalid[mode_type]],
                               sol_probe.taus[mode_type, :sol_probe._nvalid[mode_type]],
                               sol_probe.tauc[mode_type, :sol_probe._nvalid[mode_type]],
                               ]):

                ax.scatter(zz[idx_del].real, zz[idx_del].imag, c=c_fem[1], alpha=alpha_fem[1], **markerstyle_fem)
                # markerstyle_fem["zorder"] += 1
                axsScatter.append(ax.scatter(zz[idx].real, zz[idx].imag, c=range(idx.size)[::-1], alpha=alpha_fem[0], **markerstyle_fem))

            # extract the field profile of selected modes
            axs_names = ["kappa", "taus", "tauc"]
            axs_select = np.asarray(axsScatter)

            # plt.close(1) # close the figure for plotting the dispersion curve
            ####  interactively select a mode by clicking at the nearest point in the complex plane
            mode_selector = slab.SlabWaveGuide.modeSelector(
                axs_select, sol_probe, axs_names, mode_type=mode_type, nmodes = 4,
                axSpectra=axSpectra, axs_profile=np.array([ax_hz1, ax_hz2, ax_hz3, ax_hz4]))

            # plot selected modes
            # mode_selector.mpl_connect()
            mode_selector.plotModes(idx[3], ms = 8, component="h_z")
            mode_selector.plotModes(idx[2], ms = 4, component="h_z")
            mode_selector.plotModes(idx[1], ms = 8, component="h_z")
            mode_selector.plotModes(idx[0], ms = 8, component="h_z")

            kappa2_max = np.max(np.abs(np.asarray(sol_probe.n2list)))
            kappa_max = kappa2_max**0.5

            zfx = np.array([3.25, 0.95, 0.95])
            zfy = np.array([0.55, 3.55, 3.35])

            tau_regions = [["", "", ""], ["", "B", "A"], ["", "C", "D"]]
            tau_layer = ["", "s", "c"]
            n2clad = [0, sol_probe.n2list[0], sol_probe.n2list[-1]]
            ioffset = 5

            for i in range(0, 3):

                axs[i+ioffset].axline((0, 0), slope=0, linestyle="--", linewidth=1, c="k")
                axs[i+ioffset].axline((0, 0), (0,1), linestyle="--", linewidth=1, c="k")

                axs[i+ioffset].set_xlim([-zfx[i], zfx[i]])
                axs[i+ioffset].set_ylim([-zfy[i], zfy[i]])

                # write labels for tau regions

                deltax, deltay = 0.25, 0.15
                if i==0:
                    axs[i+ioffset].text(0.03, 0.08, "$(\\kappa', \\kappa'')$", usetex=True, transform=axs[i+ioffset].transAxes, fontsize=fontsize*1.25, ha="left", va="center")
                if i>0:
                    # draw a line where S_x = 0
                    c = "C"+str(int((-np.sign(n2clad[i].real)+1)/2)) # dielectric: C0; metal: C1
                    if np.abs(n2clad[i].real)<1e-12:
                        xSx0 = 0
                    else:
                        xSx0 = -n2clad[i].imag/n2clad[i].real

                    axs[i+ioffset].axline((0, 0), (xSx0, 1), linestyle="--", linewidth=1, c=c)
                    axs[i+ioffset].text(xSx0*zfy[i], zfy[i]*1.1,
                        "$\\bar{S}_x^{\\mathrm{(TM)}}=0$",
                        color=c, ha="center", va="center", fontsize=fontsize)

                    axs[i+ioffset].text(0.03, 0.08, "$(\\tau_{\mathrm "+tau_layer[i]+"}', \\tau_{\mathrm "+tau_layer[i]+"}'')$", usetex=True, transform=axs[i+ioffset].transAxes, fontsize=fontsize*1.25, ha="left", va="center")

                    for iSx in [+1, -1]:
                        for itau_im in [+1, -1]:
                            sign_n2 = np.sign(n2clad[i].real)
                            sign_Sx = int(iSx*sign_n2)
                            axs[i+ioffset].text(
                                0.5+itau_im*xSx0 + deltax*iSx, 0.5 + itau_im*deltay,
                                "$\\mathrm{"+tau_regions[(i*2-3)*itau_im][(i*2-3)*sign_Sx]+"}_\mathrm{"+tau_layer[i]+"}$",
                                transform=axs[i+ioffset].transAxes,
                                color="k", ha="center", va="center", fontsize=fontsize*1.25,
                                )


    t1 = time.process_time()
    t_CPU = t1 - t0

    print('CPU Execution time:', t_CPU, 'seconds')
    # sys.exit(0)
    #%% annotations for the figure
    # c_hz = ["C0", "C1", "C3", "C9"]
    # for i in range(0, 4):

    #     axs[i+1].text(0.25, 1.2, "\\textbf{---} $\\bar{S}_y$", usetex=True, transform=axs[i+1].transAxes, fontsize=fontsize*1.25, color="C2", ha="center", va="center", fontweight="bold")
    #     axs[i+1].text(0.75, 1.2, "$\\cdot\\cdot\\cdot\\bar{S}_x$", usetex=True, transform=axs[i+1].transAxes, fontsize=fontsize*1.25, color="C4", ha="center", va="center", fontweight="bold")
    #     axs[i+1].text(-0.10, 0.5, "$h_z$", usetex=True, transform=axs[i+1].transAxes, fontsize=fontsize, color=c_hz[i], ha="center", va="center", rotation=90)
    #     axs[i+1].text(0.5, -0.40, "x (nm)", usetex=True, transform=axs[i+1].transAxes, fontsize=fontsize*1.25, ha="center", va="top")

    #     axs[i+1].text(0.55, 0.15, "$\\mathrm{TM_"+str(i)+"}$", usetex=True, transform=axs[i+1].transAxes, fontsize=fontsize, ha="center", va="center", fontweight="bold", color=c_hz[i])

    axSpectra.text(0.6, -0.15, "$\\beta'$", usetex=True, transform=axSpectra.transAxes, fontsize=fontsize*1.25, ha="center", va="top")
    axSpectra.set_ylabel("$\\frac{\omega}{\omega_\mathrm{p}}$", rotation=0, fontsize=fontsize*1.25)
    axSpectra.set_ylim([0.09, 0.51])
    axSpectra.set_yticks([0.1, 0.3, 0.5])
    # axSpectra.set_xticks([0, 30, 60])
    # axSpectra.set_xlim([0, 55])

    # continue
    sys.exit(0)

    #%% save the figure
    figname = wg["name"]
    # plt.savefig(figname + '.png',format='png', dpi=300)
    plt.savefig(figname + '.pdf',format='pdf')
    # from PIL import Image
    # Image.open(figname+'.png').convert('L').save(figname+'-bw.png')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: shelvon
@email: xiaorun.zang@outlook.com

"""

import os
import sys
import time
import numpy as np
import matplotlib.pyplot as plt

import copy

# For a local debugging, we import or install the local package.
#
# - Method 1, append the folder's path of the local package:

try:
    import ngslab as slab
except:
    srcpath = os.path.dirname(
        os.path.dirname(
            os.path.abspath(os.getcwd())
            )
        )
    sys.path.append(srcpath)
    import ngslab as slab
#
# - Method 2, pip install locally the python package, where "." is intended
# to be run under the root folder of the python package.
# python3 -m pip install --upgrade .
#
# - Method 3, conda install the local package build by conda-build.
# conda install path/to/local/package

# import slab

# print(dir()) # print what have been imported

# sys.exit(0)
#%% 4-layer waveguide:
# a waveguide given in Fig. 2 of Ref. [J. Chilwell and I. Hodgkinson, ‘Thin-films field-transfer matrix theory of planar multilayer waveguides and reflection from prism-loaded waveguides’, J. Opt. Soc. Am. A, vol. 1, no. 7, pp. 742–753, Jul. 1984, doi: 10.1364/JOSAA.1.000742]
def wg_4layer(geom_tSlab):
    nSlab = np.maximum(geom_nPML, round(np.ceil(geom_tSlab/geom_hMax)))

    wg = dict(
        # geometry
        name = "demo_4layer",
        intervals = [-4*geom_tSlab-geom_tSub-geom_tPML, -4*geom_tSlab-geom_tSub, -4*geom_tSlab, -3*geom_tSlab, -2*geom_tSlab, -geom_tSlab, 0, geom_tFree, geom_tFree+geom_tPML],
        labels = ["pml_left", "sub", "slab4","slab3","slab2","slab1", "air", "pml_right"],
        nnodes = [geom_nPML, round(np.ceil(geom_tSub/geom_hMax)), nSlab, nSlab, nSlab, nSlab, round(np.ceil(geom_tFree/geom_hMax)), geom_nPML],
        # material
        nk = {"default":1.0, "nc": 1.0, "ns": 1.5, "n1": 1.66, "n2": 1.53, "n3": 1.60, "n4": 1.66},
        map = {"pml_left": "ns", "sub":"ns", "slab4":"n4", "slab3":"n3", "slab2":"n2", "slab1":"n1", "air":"nc", "pml_right":"nc"},
        )

    return wg

#### 4-layer lossy waveguide:
def wg_4layer_lossy(geom_tSlab):
    wg = wg_4layer(geom_tSlab);
    wg.update(
        name="demo_4layer_lossy",
        nk={"default":1.0, "nc": 1.0, "ns": 1.5, "n1": 1.66*(1+.0001j), "n2": 1.53*(1+.0001j), "n3": 1.60, "n4": 1.66},
        # nk={"default":1.0, "nc": 1+0.0j, "ns": 1.5+0.5j, "n1": 1.66*(1+0.000j), "n2": 1.53*(1+0.0000j), "n3": 1.60, "n4": 1.66},
        map = {"pml_left": "ns", "sub":"ns", "slab4":"n4", "slab3":"n3", "slab2":"n2", "slab1":"n1", "air":"nc", "pml_right":"nc"},
        )

    return wg

#### ARROW waveguide:
# in Fig. 2 of Ref. [H. P. Uranus, H. J. W. M. Hoekstra, and E. Van Groesen, ‘Simple high-order Galerkin finite element scheme for the investigation of both guided and leaky modes in anisotropic planar waveguides’, Optical and Quantum Electronics, vol. 36, no. 1, pp. 239–257, Jan. 2004, doi: 10.1023/B:OQEL.0000015643.52433.f0.]
def wg_ARROW():
    d1 = 2.0985e-6*1e6;
    d2 = 0.1019e-6*1e6;
    d3 = 4e-6*1e6;
    nSlab = np.maximum(geom_nPML, round(np.ceil(geom_tSlab/geom_hMax)));
    wg = dict(
        name = "demo_ARROW",
        intervals = [-d1-d2-d3-geom_tSub-geom_tPML, -d1-d2-d3-geom_tSub, -d1-d2-d3, -d2-d3, -d3, 0, geom_tFree, geom_tFree+geom_tPML],
        labels = ["pml_left", "sub", "slab1","slab2","slab3", "air", "pml_right"],
        nnodes = [geom_nPML, round(np.ceil(geom_tSub/geom_hMax)), nSlab, nSlab, nSlab, round(np.ceil(geom_tFree/geom_hMax)), geom_nPML],
        nk = {"default":1.0, "nc": 1.0, "ns": 3.5, "n1": 1.45, "n2": 3.5, "n3": 1.45},
        map = {"pml_left": "ns", "sub":"ns", "slab1":"n1", "slab2":"n2", "slab3":"n3", "air":"nc", "pml_right":"nc"},
        )

    return wg

#%% model creation
#---- Geometric Parameters
model = slab.SlabWaveGuide();
model.ld0_target = 6328e-10

# use micrometer may yield lower condition number
# model.ld0_target *= 1e6

geom_tPML = model.ld0_target*1.0
geom_tFree = model.ld0_target*1.5
geom_tSub = model.ld0_target*1.5

# geom_hMax = model.ld0_target/31; geom_nPML = 13; # extra fine mesh
# geom_hMax = model.ld0_target/31; geom_nPML = 27; # extra fine PML mesh
geom_hMax = model.ld0_target/17; geom_nPML = 17; # moderate resolution
# geom_hMax = model.ld0_target/17; geom_nPML = 1; # TBC test
# geom_hMax = model.ld0_target/31; geom_nPML = 1; # TBC debugging

# pml_params=[2, 1]
# pml_params=[2, 100]
# pml_params=[11, 100]
# pml_params=[101, 100]
# pml_params=[2, 10]
# pml_params=[0, 10]


pml_params=[11, 10]
# pml_params=[-1, 10]

geom_delPML = geom_tPML/geom_nPML
print("hMax="+f"{geom_hMax/model.ld0_target:.4f}*ld0")
print("hPML="+f"{geom_delPML/model.ld0_target:.4f}*ld0")

#%% loop in the slab thickness
tSlabArray = np.arange(500e-9, 501e-9, 20e-9)

plt.close("all")
# A change of unit to micrometer.
# tSlabArray *= 1e6
for it, geom_tSlab in zip(range(tSlabArray.size), tSlabArray):
    print("tSlab["+str(it)+"] = "+f"{geom_tSlab*1e3:.0f} nm")

    # choose a waveguide
    wg = wg_4layer(geom_tSlab)
    # wg = wg_4layer_lossy(geom_tSlab)
    # wg = wg_ARROW()

    # update the waveguide geometry
    model.geom.Update(wg["intervals"], wg["nnodes"], wg["labels"])

    #-- create the mesh
    model.mesh.Create()
    # model.mesh.Plot(); sys.exit(0);

    model.material.nk = wg["nk"]
    model.material.map = wg["map"]

    #%% method 1: real-time PML calculations
    # enable PML domains
    model.material.pml = {"pml_left":True, "pml_right":True}

    # model.mesh.Plot(); # sys.exit(0);

    #### model build
    #---- build the model (create function space, set up materials, enable pml materials)
    # bspl_order: the order of the ngsolve.BSpline representations for dispersive and tabulated materials
    # bspl_order = 1 (linear), 2 (quadratic) curve
    model.Build(fes_order=2, bspl_order=2, ld_scale=1, pml_params=pml_params)

    #### model simulation

    # model.w = 1.0 # by default model.w == 1.0
    t0 = time.process_time()
    sol_PML = model.Solve(show_pattern=False)
    t1 = time.process_time()
    t_CPU = t1 - t0

    print('CPU Execution time:', t_CPU, 'seconds')

    # sys.exit(0)
    n2list = np.asarray(sol_PML.n2list)
    nlist = np.asarray(n2list)**0.5
    n2delta = n2list[-1] - n2list[0]
    n2Sigma = n2list[-1] + n2list[0]
    ns, nc = nlist[[0, -1]]
    n2s, n2c = nlist[[0, -1]]**2

    #### filtering modes
    mode_markers = ["o", "^"]
    Q_threshold = 1
    # Q_threshold = (1/np.e)
    # Q_threshold = 1e-1
    kappa_tol = 1e-12
    kappa2_limit = np.max(np.abs(n2list))*100.1

    # the last one will be plotted in the complex planes
    # mode_type == 0: TM mode; mode_type == 1: TE mode
    for mode_type in [0]:
    # for mode_type in [1, 0]:

        ## results from TBC simulations
        kappa_PML = sol_PML.neff[mode_type,:]
        kappa2_PML = kappa_PML**2

        idx_kappa_PML = np.where(
            ~np.isnan(kappa_PML)
            & (np.abs(kappa_PML.real/kappa_PML.imag)>Q_threshold)
            & (np.abs(kappa2_PML)<kappa2_limit)
            )[0]

        # print(kappa_PML[idx_kappa_PML])

    # sys.exit(0)
    #%% method 2: real-time TBC calculations
    # disable PML
    model.material.pml = {"pml_left":False, "pml_right":False} # False by default

    # and then set the transparent boundary condition
    # setTBC() accepts two numbers defining the indices of the leftmost to the rightmost semi-infinite regions that bounding the core waveguide region
    model.setTBC(tbc_doms=[1, -2]) # by default: tbc_doms=[1, -2]
    # model.setTBC(tbc_doms=[0, -1]) # test

    # model.mesh.Plot(); # sys.exit(0);

    # model.TBC_PEP = False

    # sys.exit(0)

    ##### model build
    #---- build the model (create function space, set up materials, enable pml materials)
    # bspl_order: the order of the ngsolve.BSpline representations for dispersive and tabulated materials
    model.Build(fes_order=2, bspl_order=2, ld_scale=1) # bspl_order = 1 (linear), 2 (quadratic) curve

    #### model simulation

    t0 = time.process_time()
    sol_TBC = model.Solve(show_pattern=False)
    t1 = time.process_time()
    t_CPU = t1 - t0

    print('CPU Execution time:', t_CPU, 'seconds')

    # the last one will be plotted in the complex planes
    # mode_type == 0: TM mode; mode_type == 1: TE mode
    for mode_type in [0]:
    # for mode_type in [1, 0]:

        ## results from TBC simulations
        nvalid = sol_TBC._nvalid[mode_type]
        kappa_TBC = sol_TBC.kappa[mode_type, :nvalid]
        kappa2_TBC = sol_TBC.kappa2[mode_type, :nvalid]
        taus_TBC = sol_TBC.taus[mode_type, :nvalid]
        tauc_TBC = sol_TBC.tauc[mode_type, :nvalid]

        idx_kappa_TBC = np.where(
            ~np.isnan(kappa_TBC)
            & (np.abs(kappa_TBC.real/kappa_TBC.imag)>Q_threshold)
            & (np.abs(kappa2_TBC)<kappa2_limit)
            )[0]

        # divided tau plane into regions by the sign of Sx
        if mode_type==0: # TM mode
            Sx_s = (taus_TBC.real*n2s.real+taus_TBC.imag*n2s.imag)
            Sx_c = (tauc_TBC.real*n2c.real+tauc_TBC.imag*n2c.imag)
        elif mode_type==1: # TE mode for nonmagnetic material with (mu.real = 1, mu.imag=0)
            Sx_s = -(taus_TBC.real)
            Sx_c = -(tauc_TBC.real)

        # Keep modes that do not have energy feeding into the waveguide region from the infinity
        # That means, the situations of both the power flow direction points into the waveguide
        # and density increases away from the waveguide cannot be met simultaneously.
        #
        #
        #FIXME: According to the following paper
            # [1] T. Tamir and F. Kou, ‘Varieties of leaky waves and their excitation along multilayered structures’, IEEE Journal of Quantum Electronics, vol. 22, no. 4, pp. 544–551, Apr. 1986, doi: 10.1109/JQE.1986.1072991.
            # The contra-leaky mode V cannot be excited by any realistic sources.
            # However, this argument is not so obvious without giving any further explanation in that paper.
        idx_s = np.where(
            ( ~((Sx_s>-kappa_tol) & (taus_TBC.imag>-kappa_tol)) )
            )[0]
        idx_c = np.where(
            ( ~((Sx_c<kappa_tol) & (tauc_TBC.imag<kappa_tol)) )
            )[0]

        # mode indices of power flux from the denser cladding layer to the rarer cladding layer
        if ns.real>=nc.real:
            idx_denser2rarer = np.where(
                (~(
                    ((Sx_s>-kappa_tol) & (taus_TBC.imag<kappa_tol)) # power flow type As
                    &((Sx_c>-kappa_tol) & (tauc_TBC.imag<kappa_tol)) # power flow type Cc
                   ))
                )[0]

        idx = np.intersect1d(idx_s, idx_c)
        idx = np.intersect1d(idx, idx_denser2rarer)
        idx = np.intersect1d(idx_kappa_TBC, idx)
        # idx = idx_denser2rarer

        # idx = idx_kappa_fem
        idx_del = np.setdiff1d(np.arange(nvalid), idx)

    #%% method3: load t-matrix calculations in the literature
        if wg["name"]=="demo_4layer":
            neff_tmatrix = np.genfromtxt("4layer_tmatrix.txt", dtype=np.complex128)
            neff_tmatrix = neff_tmatrix.reshape((2, int(neff_tmatrix.size/2)))
        elif wg["name"]=="demo_4layer_lossy":
            neff_tmatrix = np.genfromtxt("4layer_lossy_tmatrix.txt", dtype=np.complex128)
            neff_tmatrix = neff_tmatrix.reshape((2, int(neff_tmatrix.size/2)))
        else:
            neff_tmatrix = np.full((2,1), np.nan)

        nmodes_plot = np.min([12, np.size(neff_tmatrix, 1)])
        kappa_tmatrix = neff_tmatrix[mode_type,:nmodes_plot]
        kappa2_tmatrix = kappa_tmatrix**2

        # plus a negligible imag to get the correct sign when it is pure real
        tau2_s_tmatrix = np.array(n2list[0]-kappa2_tmatrix) - 1j*1e-32
        tau2_c_tmatrix = np.array(n2list[-1]-kappa2_tmatrix) - 1j*1e-32

        taus_tmatrix = -np.sqrt(tau2_s_tmatrix)*np.sign(np.real(tau2_s_tmatrix))
        tauc_tmatrix = np.sqrt(tau2_c_tmatrix)*np.sign(np.real(tau2_c_tmatrix))

        eig_tmatrix = 0.5*(tauc_tmatrix - taus_tmatrix)


    #%% plotting the comparison
    # figure properties
    # plt.close("all")

    inch2cm = 1.0/2.54  # inches-->centimeters
    figsize = 1.0*np.array([8.6, 9.4])*inch2cm
    ms = 8
    fontsize = 8
    plt.rc('font', size=fontsize)

    fig, axs = plt.subplots(figsize=figsize, ncols=1, nrows=2, constrained_layout=True, sharex=True, sharey=False, clip_on=True);
    ll_plotted = False

    # sys.exit(0)
    #### investigate the eigenvalues
    cmaps_r = ["C7", "C3", "C2", "C5"] # complementary color of each
    c_fem = ["g", "C7"]
    alpha_fem = [1.0, 0.6]
    c_tmatrix = ["C3"]

    markerstyle_tmatrix = dict(marker="x", s=ms*4, linewidths=1.0, alpha=1.0, zorder=1e7, clip_on=True)
    markerstyle_fem = dict(marker="o", s=round(ms*1.5), edgecolor="none", zorder=1e6, clip_on=True)

    kappa_PML_selected = kappa_PML[idx_kappa_PML]
    kappa_TBC_selected = kappa_TBC[idx_kappa_TBC]
    idx_TBC_closest, idx_PML_closest = [], []
    for kappa_this in kappa_tmatrix:
        idx_TBC_closest.append(np.abs(kappa_TBC_selected-kappa_this).argmin())
        idx_PML_closest.append(np.abs(kappa_PML_selected-kappa_this).argmin())
        # print(idx_TBC_closest[-1], idx_PML_closest[-1])

    yerr_TBC = (kappa_tmatrix.imag-kappa_TBC_selected[idx_TBC_closest].imag)/np.abs(kappa_tmatrix)
    xerr_TBC = (kappa_tmatrix.real-kappa_TBC_selected[idx_TBC_closest].real)/np.abs(kappa_tmatrix)

    yerr_PML = (kappa_tmatrix.imag-kappa_PML_selected[idx_PML_closest].imag)/np.abs(kappa_tmatrix)
    xerr_PML = (kappa_tmatrix.real-kappa_PML_selected[idx_PML_closest].real)/np.abs(kappa_tmatrix)
    x, y = np.arange(kappa_tmatrix.size), np.full(np.shape(kappa_tmatrix), 0)
    x = kappa_tmatrix
    # compare modes by PML-FEM and TBC-FEM
    if idx.size>0:
        # axs[0].errorbar(
        #     x, y,
        #     yerr = 1/(-np.log10(np.abs(yerr_TBC)))**2,
        #     # xerr = 1/(-np.log10(np.abs(xerr_TBC)))**2,
        #     fmt="none", ecolor = "C3",
        #     capsize=4,
        #     )
        # axs[0].errorbar(
        #     x, y,
        #     yerr = 1/(-np.log10(np.abs(yerr_PML)))**2,
        #     # xerr = 1/(-np.log10(np.abs(xerr_PML)))**2,
        #     fmt="none", ecolor = "C7",
        #     capsize=2,
        #     )

        axs[0].plot(x, (np.abs(xerr_TBC)), 'C3-o', ms=3, lw=1.0, label="|Re($\\kappa_{\mathrm{TBC}}-\\kappa_{\mathrm{ref}}$)|")
        axs[0].plot(x, (np.abs(yerr_TBC)), 'C3--<', ms=3, lw=1.0, label="|Im($\\kappa_{\mathrm{TBC}}-\\kappa_{\mathrm{ref}}$)|")
        axs[0].plot(x, (np.abs(xerr_PML)), 'C2-o', ms=3, lw=1.0, label="|Re($\\kappa_{\mathrm{PML}}-\\kappa_{\mathrm{ref}}$)|")
        axs[0].plot(x, (np.abs(yerr_PML)), 'C2--<', ms=3, lw=1.0, label="|Im($\\kappa_{\mathrm{PML}}-\\kappa_{\mathrm{ref}}$)|")

        axs[0].set_yscale('log')
        axs[0].axvline(nc, ls="--", c="k")
        axs[0].axvline(ns, ls="--", c="k")

        axs[0].legend(handlelength=2)
    # sys.exit(0)
    # compare modes by PML-FEM and TBC-FEM
    if idx.size>0:
        axs[1].scatter(kappa_tmatrix.real, kappa_tmatrix.imag, c="C7", marker="x", s=ms*4, linewidths=2.0, alpha=1.0, clip_on=True, label="T-matrix")
        axs[1].scatter(
            np.abs(kappa_PML[idx_kappa_PML].real), kappa_PML[idx_kappa_PML].imag,
            s=ms*2,
            c="C2",
            marker="<",
            alpha=np.exp(-np.abs(kappa2_PML[idx_kappa_PML].imag)-kappa_tol)**0,
            label="PML-FEM",
            )

        axs[1].scatter(
            np.abs(kappa_TBC[idx].real), kappa_TBC[idx].imag,
            s=ms,
            # c=range(idx.size)[::-1],
            c="C3",
            # cmap="inferno",
            # alpha=np.exp(-np.abs(kappa2_TBC[idx].imag)-kappa_tol)**0,
            label="TBC-FEM",
            )
        axs[1].legend(handlelength=4, )
        axs[1].axvline(nc, ls="--", c="k")
        axs[1].axvline(ns, ls="--", c="k")

    axs[1].set_xlim([0.3, 1.65])
    axs[1].set_xlabel("$\\kappa '$")
    axs[1].set_ylabel("$\\kappa ''$")

    axs[0].set_ylim([1e-18, 1])
    axs[1].set_ylim([-0.05, 0.45])

    # axs[0].set_xlabel("$\\kappa '$")
    # axs[0].set_ylabel("$|\\Delta \\kappa| = |\\kappa - \\kappa_{\mathrm{ref}}|$")
    axs[0].set_ylabel("relative difference")

    for i in range(2):
        axs[i].text(-0.05, 1.05, '('+chr(i+97)+')', transform=axs[i].transAxes, fontsize=fontsize*1.25, ha="right", va="bottom", fontweight="book")


    axs[0].set_title("$s' + i s'' = (1 "+f"{pml_params[0]-1:+G}"+"|\\frac{\\Delta x}{L}|^2) + "+str(pml_params[1])+"i |\\frac{\\Delta x}{L}|^2$")


    plt.legend(framealpha=1.0)
    # sys.exit(0)
    # continue
    #%% save the figure
    # figname = wg["name"]+"pml_1"
    # figname = wg["name"]+"pml_2"
    figname = wg["name"]+"pml_params_"+str(pml_params[0]-1)+"_"+str(pml_params[1])
    # plt.savefig(figname + '.png',format='png', dpi=300)
    plt.savefig(figname + '.pdf',format='pdf')
    # from PIL import Image
    # Image.open(figname+'.png').convert('L').save(figname+'-bw.png')
